#include <tiuser.h>
#include <stropts.h>
#include <stdio.h>
#include <fcntl.h>
#include <signal.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>

extern int t_errno;

#define LOOPADDR "example.addr"


#define SERV_PORT 5134
#define MAX_CONN  50

struct t_call *call[MAX_CONN];

main()
{
   int listen_fd;       /* file descriptor into transport */
   int recfd;           /* fd on which to accept          */
   struct t_bind *req;  /* requested addr to which to bind*/
   struct t_bind *ret;  /* actual address that was bound  */
   int i, j, curr_call; /* index counters                 */
   int new_listen;      /* flag for new connection        */
   struct t_discon *discon;   /* holds disconnect info    */
   struct sockaddr_in myaddr; /* holds the address of     */
                              /* this service             */

   if ((listen_fd = t_open("/dev/ticots", O_RDWR, NULL))
                                                      < 0) {
      t_error("t_open failed");
      exit(1);
   }

   /*
    *   Get space to hold the address to which
    *   you want to bind.
    */

   if ((req = (struct t_bind *)t_alloc(listen_fd, T_BIND,
                                T_ALL)) == NULL) {
      t_error("t_alloc failed");
      exit(1);
   }

   /*
    *   Get space to hold the address to which
    *   you actually bind.
    */

   if ((ret = (struct t_bind *)t_alloc(listen_fd, T_BIND,
                               T_ALL)) == NULL) {
      t_error("t_alloc failed");
      exit(1);
   }
   /*
    *  Set qlen to MAX_CONN to signify that we will let up
    *  to MAX_CONN more connection requests to arrive
    *  simultaneously.
    */

   req->qlen = MAX_CONN;

   req->addr.len = strlen(LOOPADDR) + 1;
   strncpy(req->addr.buf, LOOPADDR, strlen(LOOPADDR) + 1);

   /*
    *  Bind to the address to which the service will
    *  be offered.
    */

   if (t_bind(listen_fd, req, ret) < 0) {
      t_error("t_bind failed");
      exit(1);
   }

   /*
    *  Make sure the address you got was the one you wanted!
    */

   if ((req->addr.len != ret->addr.len) ||
     (memcmp(req->addr.buf, ret->addr.buf, req->addr.len))){
      fprintf(stderr, "Did not bind to correct address!\\n");
      exit(1);
   }

   /*
    *  Loop continuously, listening for connection requests,
    *  accepting the call, and performing the service.
    */

   while (1) {
      /*
       *  First, find an unused connection slot.
       */

      for (curr_call=0; curr_call < MAX_CONN; curr_call++) {
         if (call[curr_call] == NULL) {
            break;
         }
      }

      /*
       *  Allocate space to hold the t_call structure, and
       *  listen for incoming calls.  The t_listen() routine
       *  sleeps until a call arrives (it returns
       *  immediately if a call is already present).
       */

      if ((call[curr_call] =
           (struct t_call *)t_alloc(listen_fd, T_CALL,
                                    T_ALL)) == NULL) {
         t_error("could not allocate space for t_call!");
         exit(1);
      }

      if (t_listen(listen_fd, call[curr_call]) < 0) {
         t_error("could not listen!");
         exit(1);
      }

      /*
       *  Now that we have received a call,  let's try to
       *  accept it.  The t_accept will fail if another call
       *  arrives.   In that case, we return to the loop to
       *  receive the new call.  So, at this point, we may
       *  have several calls waiting to be accepted.
       *  Therefore, we loop through all call structures and
       *  accept all of them.
       */

      for (i = 0; i < MAX_CONN; i++) {
         if (call[i] == NULL) {
            continue;
         }

         if ((recfd = t_open("/dev/ticots", O_RDWR, NULL))
                                                      < 0) {
            t_error("could not open device!");
            exit(1);
         }

         if (t_bind(recfd, NULL, NULL) < 0) {
            t_error("bind for responding failed!");
            exit(1);
         }

         new_listen = 0;
         if (t_accept(listen_fd, recfd, call[i]) < 0) {
            if (t_errno != TLOOK) {
               t_error("t_accept failed");
               exit(1);
            } else {
               /*
                *  Figure out why t_accept() failed.
                */

               switch (t_look(listen_fd)) {

                  case T_DISCONNECT:

                     /*
                      *  A disconnect happened on one of the
                      *  incoming calls.
                      */
                     if ((discon = (struct t_discon *)
                        t_alloc(listen_fd, T_DIS, T_ALL))
                                                  == NULL) {
                        t_error("t_alloc failed");
                        exit(1);
                     }
                     if (t_rcvdis(listen_fd, discon) <0) {
                        t_error("could not disconnect");
                        exit(1);
                     }

                     /*
                      *  Figure out which call disconnected.
                      */

                     for (j = 0; j < MAX_CONN; j++) {
                        if (discon->sequence ==
                                        call[j]->sequence) {
                           t_free(call[j], T_CALL);
                           call[j] = NULL;
                           break;
                        }
                     }
                     t_free(discon, T_DIS);
                     break;

                  case T_LISTEN:

                     new_listen = 1;
                     break;

                  default:

                     t_error("t_accept failed");
                     exit(1);
               }
            }

            /*
             *  If t_accept() failed because a new
             *  connection indication arrived, break
             *  the for loop.  That causes us to go back
             *  to the while loop and call t_listen()
             *  again.  Otherwise, we failed because we got
             *  a disconnect.  Because we've cleared that,
             *  continue the for loop to try to accept the
             *  next connection.
             */

            if (new_listen) {
               break;
            } else {
               continue;
            }
         }

         /*
          *  At this point, we have an accepted connection.
          *  So, we free the call structure and implement
          *  the protocol.
          */

         t_free(call[i], T_CALL);
         call[i] = NULL;

         switch (fork()) {
            case -1:
               perror("fork failed!\\n");
               exit(1);
   
            default: /* parent */
   
               t_close(recfd);
               break;
   
            case 0:  /* child */
   
               t_close(listen_fd);
   
               /*
                *  The child_actions() routine exits
                *  when it completes.
                */
               child_actions(recfd);
         }
      }
   }
}
