#include <tiuser.h>
#include <stropts.h>
#include <stdio.h>
#include <fcntl.h>
#include <signal.h>
#include <sys/types.h>
#include <sys/poll.h>
#include <shadow.h>
#include <pwd.h>
#include "tpinfo.h"

extern int t_errno;

void process_it();

#define MAX_DATA    9999
#define HDR_SIZE    4

#define BSIZE  MAX_DATA + HDR_SIZE

main()
{
   struct t_bind *req;  /* address to which to bind       */
   struct t_bind *ret;  /* actual address that was bound  */

   struct pollfd  *pollfdp;        /* holds all endpoints */
   struct pollfd  *curr_pollfdp;   /* traverses pollfdp   */

   char     **devicepp;     /* array of device names      */
   char   **curr_devicepp;  /* traverses devicepp         */
   struct tpinfo  *tpinfop; /* return value of get_addrs()*/

   struct tpinfo  *curr_tpinfop;    /* traverses tpinfop  */

   int    num_addrs;  /* number of address for a provider */
   int    num_fd;     /* number of endpoints to use       */
   int    fd;         /* holds current endpoint           */
   int    i;          /* loop index                       */

   /*
    *  Get all addresses and transport providers on the
    *  system...
    */

   if ((tpinfop = get_addrs(NULL, "rpopen")) == NULL) {
      fprintf(stderr,
             "There are no addresses for this service!\n");
           exit(1);
   }
   /*
    *  Count the number of transport endpoints we'll need.
    */

   curr_tpinfop = tpinfop;
   num_fd = 0;
   while (curr_tpinfop) {
      num_fd += curr_tpinfop->nd_addrlistp->n_cnt;
      curr_tpinfop = curr_tpinfop->nextp;
   }
   /*
    *  Allocate space to hold the information about
    *  each transport endpoint.
    */

   if ((pollfdp =
    (struct pollfd *)malloc(num_fd * sizeof(struct pollfd)))
                                                  == NULL) {
         fprintf(stderr,
             "Cannot allocate space for file descriptors\n");
         exit(1);
   }

   if ((devicepp =
      (char **)malloc(num_fd * sizeof(char *))) == NULL) {
         fprintf(stderr,
                "Cannot allocate space for device names\n");
         exit(1);
   }

   /*
    *  Create a bound transport endpoint for each address
    *  returned from get_addrs(), fill in the pollfd
    *  structure for each endpoint, and store the name
    *  of each device file.
    */

   curr_pollfdp = pollfdp;
   curr_devicepp = devicepp;
   curr_tpinfop = tpinfop;
   while (curr_tpinfop) {
      num_addrs = curr_tpinfop->nd_addrlistp->n_cnt;
      while (num_addrs --) {
         if ((curr_pollfdp->fd= t_open(curr_tpinfop->device,
                                       O_RDWR, NULL)) < 0) {
            t_error("t_open failed");
            exit(1);
         }
         curr_pollfdp->events = POLLIN;
         *curr_devicepp = curr_tpinfop->device;
         fd = curr_pollfdp->fd;

         if ((req = (struct t_bind *)t_alloc(fd, T_BIND,
                                             0)) == NULL) {
            t_error("t_alloc failed");
            exit(1);
         }

         if ((ret = (struct t_bind *)t_alloc(fd, T_BIND,
                                         T_ALL)) == NULL) {
            t_error("t_alloc failed");
            exit(1);
         }

         req->qlen = 1;
         req->addr= *(curr_tpinfop->nd_addrlistp->n_addrs);

         /*
          *  Bind to the address over which the service
          *  will be offered.
          */

         if (t_bind(fd, req, ret) < 0) {
            t_error("t_bind failed");
            exit(1);
         }

         /*
          *  Make sure the address you got was the one
          *  you wanted!
          */

         if ((req->addr.len != ret->addr.len) ||
            (memcmp(req->addr.buf, ret->addr.buf,
                                          req->addr.len))) {
            fprintf(stderr,
                      "Did not bind to correct address!\n");
            exit(1);
         }
         curr_pollfdp++;
         curr_devicepp++;
         curr_tpinfop->nd_addrlistp->n_addrs++;
      }
      curr_tpinfop = curr_tpinfop->nextp;
   }

   /*
    *  Loop forever, polling for events.  When an event
    *  occurs, find the transport endpoint and call
    *  the process_it() routine to implement the protocol.
    */

   while (1) {
      if (poll(pollfdp, num_fd, -1) < 0) {
         perror("poll failed!");
         exit(1);
      }

      curr_pollfdp = pollfdp;
      curr_devicepp = devicepp;
      for (i = 0; i < num_fd; i++) {
         switch(curr_pollfdp->revents) {
            case 0:
               break;
            case POLLIN:
               process_it(curr_pollfdp->fd, *curr_devicepp);
               break;
            default:
               perror("poll returned an error");
               exit(1);
         }
         curr_pollfdp ++;
         curr_devicepp ++;
      }
   }
}
void
process_it(fd, device)
int fd;
char *device;
{
   struct t_call *call;  /* holds connection info       */
   int recfd;            /* endpoint on which to accept */

   switch (t_look(fd)) {
      case T_DISCONNECT:
         if (t_rcvdis(fd, NULL) < 0) {
            t_error("t_rcvdis");
         }
         return;

      default:
         return;

      case T_LISTEN:
         if ((call = (struct t_call *)t_alloc(fd,
                             T_CALL, T_ALL)) == NULL) {
            t_error("could not get space for t_call!");
            exit(1);
         }
         if (t_listen(fd, call) < 0) {
            t_error("could not listen!");
            exit(1);
         }

         if ((recfd =
                   t_open(device, O_RDWR, NULL)) < 0) {
            t_error("could not open device!");
            exit(1);
         }

         if (t_bind(recfd, NULL, NULL) < 0) {
            t_error("bind for responding failed!");
            exit(1);
         }

         if (t_accept(fd, recfd, call) < 0) {
            /*
             * Did it fail because client disconnected?
             * t_errno is set to TLOOK if it did.
             */

            if (t_errno == TLOOK) {
               if (t_rcvdis(fd, NULL) < 0) {
                  t_error("could not disconnect");
                  exit(1);
               }
               (void)t_close(recfd);
               return;
            } else {
               t_error("could not accept call");
               exit(1);
            }
         }

         switch (fork()) {
            case -1:
               perror("fork failed!\n");
               exit(1);

            default: /* parent */

               (void)t_close(recfd);
               break;

            case 0:  /* child */

               (void)t_close(fd);

               /*
                *  The child_actions() routine exits
                *  when it completes.
                */
               child_actions(recfd);
         }
   }     
}

int
child_actions(conn_fd)
int conn_fd;
{
   int done;             /* indicates all data are read   */
   int where;            /* where we are in buffer        */
   int nbytes;           /* the number of bytes read      */
   int length;           /* the length indicator          */
   int flags = 0;        /* info returned from t_rcv()    */
   FILE *fp;             /* file pointer into popen()     */
   char *response;       /* used for authentication       */
   struct spwd *spent;   /* used for authentication       */
   struct passwd *pwent; /* holds user information        */
   char type;            /* "r" or "w" from client        */
   char name[BSIZE];     /* user name from client         */
   char passwd[BSIZE];   /* password from client          */
   char command[BSIZE];  /* command to give popen()       */
   char buff[BSIZE];     /* holds data from popen()       */
   char sendbuff[BSIZE]; /* hold data we send to client   */
   void hangup();        /* signal handler for disconnect */

   /*
    *  Read the user's name.
    */

   done = 0;
   where = 0;
   do {
      if ((BSIZE - where) == 0) {
          exit(1);
      }
      if ((nbytes = t_rcv(conn_fd, &name[where],
                             BSIZE - where, &flags)) < 0) {
         t_error("t_rcv of user name failed!");
         exit(1);
      }
      where += nbytes;
      if (name[where - 1] == '\0') {
         done = 1;
      }
   } while (!done);

   /*
    * Read the user's password...
    */

   done = 0;
   where = 0;
   do {
      if ((BSIZE - where) == 0) {
         exit(1);
      }
      if ((nbytes = t_rcv(conn_fd, &passwd[where],
                             BSIZE - where, &flags)) < 0) {
         t_error("t_rcv of password failed!");
         exit(1);
      }
      where += nbytes;
      if (passwd[where - 1] == '\0') {
         done = 1;
      }
   } while (!done);

   response = "y";
   if ((pwent = getpwnam(name)) == NULL
     || (spent = getspnam(name)) == NULL
     || strcmp(crypt(passwd,spent->sp_pwdp), spent->sp_pwdp)
                                                     != 0) {
         response = "n";
   }

   if (t_snd(conn_fd, response, 1, 0) < 0) {
         t_error("t_snd of password response failed!");
         exit(1);
   }

   /*
    *  If we sent an "n" character, wait for the client
    *  side of the application to disconnect.  The t_rcv()
    *  routine returns (with an error) when the disconnect
    *  occurs.
    */

   if (strcmp(response, "n") == 0) {
       (void)t_rcv(conn_fd, &type, 1, &flags);
       exit(0);
   }

   /*
    *  Set the user ID and group ID to the specified
    *  user...
    */

   if (setuid(pwent->pw_uid) < 0
      || setgid(pwent->pw_gid) < 0) {
         exit(1);
   }

   /*
    *  Read a w or an r.
    */

   if (t_rcv(conn_fd, &type, 1, &flags) < 0) {
      t_error("t_rcv of type failed!");
      exit(1);
   }
 
   /*
    *  Read the command
    */

   done = 0;
   where = 0;
   do {
      if ((BSIZE - where) == 0) {
         exit(1);
      }
      if ((nbytes = t_rcv(conn_fd, &command[where],
                             BSIZE - where, &flags)) < 0) {
         t_error("t_rcv of command failed!");
         exit(1);
      }
      where += nbytes;
      if (command[where - 1] == '\0') {
         done = 1;
      }
   } while (!done);

   if (type == 'r') {
      if ((fp = popen(command, "r")) == NULL) {
         exit(1);
      }

      /*
       *  Read from the popen() call, put in the data length,
       *  and send the message to the client.
       *  Before doing that, set up the transport endpoint to
       *  generate a SIGPOLL signal if a message appears.
       *  Because we do not expect any messages, we know a
       *  message is a disconnect indication.
       */

      signal(SIGPOLL, hangup);
      if (ioctl(conn_fd, I_SETSIG, S_INPUT) < 0) {
         perror("can't set S_INPUT");
         exit(1);
      }

      /*
       *  Check if the disconnect arrived before we set
       *  the signal catcher
       */

      if (t_look(conn_fd) != 0) {
         fprintf(stderr, "Unexpected disconnect detected\n");
         exit(1);
      }

      while ((nbytes = fread(buff, 1, BSIZE, fp)) != NULL){
         sprintf(sendbuff, "%.4d", nbytes);
         memcpy(sendbuff + HDR_SIZE, buff, nbytes);
         if (t_snd(conn_fd, sendbuff, nbytes + HDR_SIZE, 0) < 0) {
            t_error("t_snd of data failed!");
            exit(1);
         }
      }

      if (ioctl(conn_fd, I_SETSIG, 0) < 0) {
         perror("can't turn off S_INPUT");
         exit(1);
      }

      /*
       *  Send the "0000" string to indicate end-of-data and
       *  wait for the client side to disconnect.
       */

      pclose(fp);
      if (t_snd(conn_fd, "0000", HDR_SIZE, 0) != HDR_SIZE) {
         t_error("t_snd of 0 failed!");
         exit(1);
      }
      (void)t_rcv(conn_fd, &type, 1, &flags);
      exit(0);
   } else {
      if ((fp = popen(command, "w")) == NULL) {
         exit(1);
      }
      for (;;) {
         /*
          *  Get the first 4 bytes (which contain
          *  the length indicator).
          */

         where = 0;
         do {
            if ((nbytes = t_rcv(conn_fd, &buff[where],
                                HDR_SIZE - where, &flags)) < 0) {
               t_error("t_rcv of command failed!");
               exit(1);
            }
            where += nbytes;
         } while (where < HDR_SIZE);

         (void)sscanf(buff, "%4d", &length);

         /*
          *   Check if this packet is the end-of-data marker.
          *   If it is, send the disconnect request to close the
          *   connection.
          */

         if (length == 0) {
            (void)t_snddis(conn_fd, NULL);
            pclose(fp);
            exit(0);
         }

         /*
          *  Because we are not at the end-of-data marker,
          *  read the data in this packet and write
          *  them to the pipe.
          */

        where = 0;
        do {
            if ((nbytes = t_rcv(conn_fd, &buff[where],
                                length - where, &flags)) < 0) {
               t_error("t_rcv of command failed!");
               exit(1);
            }
            where += nbytes;
         } while (where < length);

         if (fwrite(buff, 1, length, fp) != length) {
            exit(1);
         }
      }
   }
}

void
hangup()
{
   fprintf(stderr, "Unexpected disconnect detected\n");
   exit(1);
}
