#include <stdio.h>
#include <netconfig.h>
#include <netdir.h>
#include <tiuser.h>

main(argc, argv)
int argc;
char *argv[];
{
    void *handlep;                    /* routine handle  */
    struct netconfig *netconfigp;     /* provider info   */
    struct nd_hostserv nd_hostserv;   /* host/serv. name */
    struct nd_addrlist *nd_addrlistp; /* addrs returned  */
    int num_addresses;                /* number of addrs */

    if (argc != 3) {
        fprintf(stderr,
                "%s: usage: %s server_name service_name\n",
                argv[0], argv[0]);
        exit(1);
    }

    /*
     *  Initialize the Network Selection routines.
     */

    if ((handlep = setnetpath()) == NULL) {
        nc_perror(argv[0]);
        exit(1);
    }

    /*
     *  Set up the information for the Name-to-Address
     *  Mapping routines.
     */

    nd_hostserv.h_host = argv[1];
    nd_hostserv.h_serv = argv[2];

    /*
     *  Loop through all visible transport providers
     *  on the system.
     */

    while ((netconfigp = getnetpath(handlep)) != NULL) {

        printf("For transport %s", netconfigp->nc_netid);

        switch(netconfigp->nc_semantics) {
            case NC_TPI_COTS:
            case NC_TPI_COTS_ORD:
                printf("(connection-oriented):");
                break;
            case NC_TPI_CLTS:
                printf("(connectionless):");
                break;
            default:
                printf("(unknown!):");
        }

        /*
         *  Get the address of the service for the
         *  current transport provider.
         */

        if (netdir_getbyname(netconfigp,
                &nd_hostserv, &nd_addrlistp) != 0) {
            num_addresses = 0;
        } else {
            num_addresses = nd_addrlistp->n_cnt;
        }

        printf(" %d address%s found\n", num_addresses,
                   num_addresses == 1? "" : "es");
    }
    endnetconfig(handlep);
}
